import numpy as np

from .strategy import Strategy


class RandomSampling(Strategy):
    def __init__(
        self,
        pool_images,
        pool_labels,
        val_images,
        val_labels,
        test_images,
        test_labels,
        idxs_lb,
        model,
        dataset,
        dataset_params,
    ):
        super(RandomSampling, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model,
            dataset,
            dataset_params,
        )

    def query(self, n):
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        np.random.shuffle(idxs_unlabeled)
        idxs_queried = idxs_unlabeled[:n]  # idxs_queried: index in pool_images
        return idxs_queried



import json
from typing import Any, List

import matplotlib.pyplot as plt


def plot_loss(
    full_result, random_result, strategy_result, save_path="./output/loss.png"
):

    random_loss = [log["dice_loss"] for log in random_result]
    strategy_loss = [log["dice_loss"] for log in strategy_result]
    full_loss = [full_result["dice_loss"]] * len(random_loss)

    plt.plot(full_loss, label="Supervised", linestyle="dashed")
    plt.plot(random_loss, label="Random", marker="o")
    plt.plot(strategy_loss, label="Strategy", marker="o")
    plt.title("Loss per round")
    plt.ylabel("DiceLoss")
    plt.xlabel("round")
    plt.legend(), plt.grid()
    plt.savefig(save_path)
    # plt.show()


def plot_score(
    full_result, random_result, strategy_result, save_path="./output/score.png"
):

    random_score = [log["iou_score"] for log in random_result]
    strategy_score = [log["iou_score"] for log in strategy_result]
    full_score = [full_result["iou_score"]] * len(random_score)

    plt.plot(full_score, label="Supervised", linestyle="dashed")
    plt.plot(random_score, label="Random", marker="o")
    plt.plot(strategy_score, label="Strategy", marker="o")
    plt.title("Score per round")
    plt.ylabel("mean IoU")
    plt.xlabel("round")
    plt.legend(), plt.grid()
    plt.savefig(save_path)
    # plt.show()


def save_json(data: Any, save_path: str) -> None:
    """Save data to json format.
    Args:
        data (Any): The data to store.
        save_path (str): Path to save.
    """
    with open(save_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, ensure_ascii=False)


def load_json(path: str) -> List[Any]:
    """Read json data
    Args:
        path (str): Path of file to read.
    Returns:
        List[Any]: FTSE entity data.
    """
    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)
    return data



from typing import List, Optional

import albumentations as albu
import cv2
import numpy as np
from albumentations import BaseCompose
from torch.utils.data import Dataset


def get_training_augmentation():
    train_transform = [
        albu.HorizontalFlip(p=0.5),
        albu.ShiftScaleRotate(
            scale_limit=0.5, rotate_limit=0, shift_limit=0.1, p=1, border_mode=0
        ),
        albu.PadIfNeeded(
            min_height=320, min_width=320, always_apply=True, border_mode=0
        ),
        albu.RandomCrop(height=320, width=320, always_apply=True),
        albu.GaussNoise(p=0.2),
        albu.Perspective(p=0.5),
        albu.OneOf(
            [
                albu.CLAHE(p=1),
                albu.RandomBrightness(p=1),
                albu.RandomGamma(p=1),
            ],
            p=0.9,
        ),
        albu.OneOf(
            [
                albu.Sharpen(p=1),
                albu.Blur(blur_limit=3, p=1),
                albu.MotionBlur(blur_limit=3, p=1),
            ],
            p=0.9,
        ),
        albu.OneOf(
            [
                albu.RandomContrast(p=1),
                albu.HueSaturationValue(p=1),
            ],
            p=0.9,
        ),
    ]
    return albu.Compose(train_transform)


def get_validation_augmentation():
    """Add paddings to make image shape divisible by 32"""
    test_transform = [albu.PadIfNeeded(384, 480)]
    return albu.Compose(test_transform)


def to_tensor(x, **kwargs):
    return x.transpose(2, 0, 1).astype("float32")


def get_preprocessing(preprocessing_fn):
    """Construct preprocessing transform

    Args:
        preprocessing_fn (callbale): data normalization function
            (can be specific for each pretrained neural network)
    Return:
        transform: albumentations.Compose

    """

    _transform = [
        albu.Lambda(image=preprocessing_fn),
        albu.Lambda(image=to_tensor, mask=to_tensor),
    ]
    return albu.Compose(_transform)


class CamvidDataset(Dataset):
    """CamVid Dataset. Read images, apply augmentation and preprocessing transformations.

    Args:
        image_paths (List[str]): path to images
        mask_paths (List[str]): path to segmentation masks
        class_values (list): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. noralization, shape manipulation, etc.)

    """

    CLASSES = [
        "sky",
        "building",
        "pole",
        "road",
        "pavement",
        "tree",
        "signsymbol",
        "fence",
        "car",
        "pedestrian",
        "bicyclist",
        "unlabelled",
    ]

    def __init__(
        self,
        image_paths: List[str],
        mask_paths: List[str],
        classes: Optional[str] = None,
        augmentation: Optional[BaseCompose] = None,
        preprocessing: Optional[BaseCompose] = None,
    ):
        self.image_paths = image_paths
        self.mask_paths = mask_paths

        # convert str names to class values on masks
        self.class_values = [self.CLASSES.index(cls.lower()) for cls in classes]

        self.augmentation = augmentation
        self.preprocessing = preprocessing

    def __getitem__(self, i):

        # read data
        image = cv2.imread(self.image_paths[i])
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask = cv2.imread(self.mask_paths[i], 0)

        # extract certain classes from mask (e.g. cars)
        masks = [(mask == v) for v in self.class_values]
        mask = np.stack(masks, axis=-1).astype("float")

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        # apply preprocessing
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def __len__(self):
        return len(self.image_paths)



import numpy as np

from .strategy import Strategy


class LeastConfidence(Strategy):
    def __init__(
        self,
        pool_images,
        pool_labels,
        val_images,
        val_labels,
        test_images,
        test_labels,
        idxs_lb,
        model,
        dataset,
        dataset_params,
    ):
        super(LeastConfidence, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model,
            dataset,
            dataset_params,
        )

    def get_topk_idxs(self, scores, k):
        if isinstance(scores, list):
            scores = np.array(scores)
        return scores.argsort()[::-1][:k]

    def query(self, n):
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        probs = self.predict_prob(idxs_unlabeled)
        scores = self.cal_scores(probs)
        topk_idxs = self.get_topk_idxs(scores, n)  # index in scores
        idxs_queried = idxs_unlabeled[topk_idxs]  # idxs_queried: index in pool_images
        return idxs_queried

    @staticmethod
    def cal_scores(probs):
        scores = []
        max_conf = np.max(probs, axis=1)
        for conf in max_conf:
            scores.append(np.mean(conf))
        scores = (
            np.array(scores) * -1
        )  # the smaller the better (confidence is low) -> Reverse it makes the larger the better
        return scores



import numpy as np

from .strategy import Strategy


class EntropySampling(Strategy):
    def __init__(
        self,
        pool_images,
        pool_labels,
        val_images,
        val_labels,
        test_images,
        test_labels,
        idxs_lb,
        model,
        dataset,
        dataset_params,
    ):
        super(EntropySampling, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model,
            dataset,
            dataset_params,
        )

    def get_topk_idxs(self, scores, k):
        return scores.argsort()[::-1][:k]

    def query(self, n):
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        probs = self.predict_prob(idxs_unlabeled)
        scores = self.cal_scores(probs)  # The larger the better
        topk_idxs = self.get_topk_idxs(scores, n)  # index in scores
        idxs_queried = idxs_unlabeled[topk_idxs]  # idxs_queried: index in pool_images
        return idxs_queried

    @staticmethod
    def cal_scores(probs):  # B,C,H,W
        scores = []
        for i in range(len(probs)):  # one img prob
            entropy = np.mean(
                -np.nansum(np.multiply(probs[i], np.log2(probs[i] + 1e-12)), axis=0)
            )
            scores.append(entropy)
        return np.array(scores)



from pathlib import Path

import numpy as np
import segmentation_models_pytorch as smp
import torch
import torch.nn.functional as F
from segmentation_models_pytorch import utils
from torch.utils.data import DataLoader


class Strategy:
    def __init__(
        self,
        pool_images,
        pool_labels,
        val_images,
        val_labels,
        test_images,
        test_labels,
        idxs_lb,
        model_params,
        dataset,
        dataset_params,
    ):
        self.pool_images = pool_images
        self.pool_labels = pool_labels
        self.val_images = val_images
        self.val_labels = val_labels
        self.test_images = test_images
        self.test_labels = test_labels
        self.idxs_lb = idxs_lb  # bool type
        self.dataset = dataset
        self.dataset_params = dataset_params  # dict
        self.val_dataset = None
        self.test_dataset = None
        self.n_pool = len(pool_images)
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model_params = model_params
        self.best_model = None
        self.train_logs = []
        self.val_logs = []
        self.test_logs = []

    def query(self, n):
        pass

    def update(self, idxs_lb):
        self.idxs_lb = idxs_lb

    def train(self, n_epoch=10, activation="softmax2d", base_path="output"):
        model = smp.__dict__[self.model_params["MODEL_NAME"]](
            encoder_name=self.model_params["ENCODER"],
            encoder_weights=self.model_params["ENCODER_WEIGHTS"],
            classes=self.model_params["NUM_CLASSES"],
        )
        model.to(self.device)

        if type(base_path) is str:
            base_path = Path(base_path)
        if not base_path.exists():
            base_path.mkdir(exist_ok=True, parents=True)

        idxs_train = np.arange(self.n_pool)[self.idxs_lb]

        train_images = [self.pool_images[idx] for idx in idxs_train]
        train_labels = [self.pool_labels[idx] for idx in idxs_train]

        train_dataset = self.dataset(
            train_images,
            train_labels,
            augmentation=self.dataset_params["training_augmentation"],
            preprocessing=self.dataset_params["preprocessing"],
            classes=self.dataset_params["classes"],
        )

        if self.val_dataset is None:
            valid_dataset = self.dataset(
                self.val_images,
                self.val_labels,
                augmentation=self.dataset_params["validation_augmentation"],
                preprocessing=self.dataset_params["preprocessing"],
                classes=self.dataset_params["classes"],
            )
            self.valid_dataset = valid_dataset
        else:
            valid_dataset = self.valid_dataset

        if self.device == "cuda":
            train_loader = DataLoader(
                train_dataset, batch_size=4, shuffle=True, num_workers=4
            )
            valid_loader = DataLoader(
                valid_dataset, batch_size=10, shuffle=False, num_workers=2
            )
        else:
            train_loader = DataLoader(train_dataset, batch_size=4, shuffle=True)
            valid_loader = DataLoader(valid_dataset, batch_size=10, shuffle=False)

        loss = utils.losses.DiceLoss(activation=activation)
        metrics = [
            utils.metrics.IoU(threshold=0.5, activation=activation),
        ]

        optimizer = torch.optim.Adam(
            [
                dict(params=model.parameters(), lr=0.0001),
            ]
        )

        train_epoch = utils.train.TrainEpoch(
            model,
            loss=loss,
            metrics=metrics,
            optimizer=optimizer,
            device=self.device,
            verbose=True,
        )

        valid_epoch = utils.train.ValidEpoch(
            model,
            loss=loss,
            metrics=metrics,
            device=self.device,
            verbose=True,
        )

        max_score = 0
        for i in range(0, n_epoch):

            print("\nEpoch: {}".format(i))
            train_logs = train_epoch.run(train_loader)
            valid_logs = valid_epoch.run(valid_loader)
            self.train_logs.append(train_logs)
            self.val_logs.append(valid_logs)

            # Save best model
            if max_score < valid_logs["iou_score"]:
                max_score = valid_logs["iou_score"]
                self.best_model = model
                model_path = base_path / "best_model.pth"
                torch.save(model, model_path)
                print("Model saved!")

        # Evaluate model on test data
        if self.test_dataset is None:
            test_dataset = self.dataset(
                self.test_images,
                self.test_labels,
                augmentation=self.dataset_params["validation_augmentation"],
                preprocessing=self.dataset_params["preprocessing"],
                classes=self.dataset_params["classes"],
            )
            self.test_dataset = test_dataset
        else:
            test_dataset = self.test_dataset

        print("Evaluate on test data")
        test_dataloader = DataLoader(test_dataset, batch_size=10, shuffle=False)
        test_epoch = utils.train.ValidEpoch(
            model=self.best_model,
            loss=loss,
            metrics=metrics,
            device=self.device,
        )

        round_test_log = test_epoch.run(test_dataloader)
        self.test_logs.append(round_test_log)

        return round_test_log

    def predict_prob(self, idxs_unlabeled):

        model = self.best_model
        unlabeled_images = [self.pool_images[idx] for idx in idxs_unlabeled]
        unlabeled_labels = [self.pool_labels[idx] for idx in idxs_unlabeled]

        unlabeled_dataset = self.dataset(
            unlabeled_images,
            unlabeled_labels,
            augmentation=self.dataset_params["validation_augmentation"],
            preprocessing=self.dataset_params["preprocessing"],
            classes=self.dataset_params["classes"],
        )

        if self.device == "cuda":
            unlabeled_loader = DataLoader(
                unlabeled_dataset, batch_size=10, shuffle=False, num_workers=2
            )
        else:
            unlabeled_loader = DataLoader(
                unlabeled_dataset, batch_size=10, shuffle=False
            )

        probs = []
        for batch_images, _ in unlabeled_loader:
            if self.device == "cuda":
                batch_images = batch_images.to(self.device)
            out = model.predict(batch_images)
            batch_probs = F.softmax(out, dim=1)
            batch_probs = batch_probs.detach().cpu().numpy()
            probs.append(batch_probs)

        probs = np.concatenate(probs, axis=0)

        return probs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['segal', 'segal.datasets', 'segal.strategies']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.3.0,<2.0.0',
 'matplotlib>=3.6.0,<4.0.0',
 'scipy>=1.8.0,<2.0.0',
 'segmentation-models-pytorch==0.3.0']

setup_kwargs = {
    'name': 'segal',
    'version': '0.1.2',
    'description': 'SegAL is an active learning freamwork for semantice segmentation.',
    'long_description': '# SegAL\n\n<p align="center">\n  <a href="https://github.com/BrambleXu/segal/actions?query=workflow%3ACI">\n    <img src="https://img.shields.io/github/workflow/status/BrambleXu/segal/CI/main?label=CI&logo=github&style=flat-square" alt="CI Status" >\n  </a>\n  <a href="https://python-poetry.org/">\n    <img src="https://img.shields.io/badge/packaging-poetry-299bd7?style=flat-square&logo=data:image/png" alt="Poetry">\n  </a>\n  <a href="https://github.com/ambv/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square" alt="black">\n  </a>\n  <a href="https://github.com/pre-commit/pre-commit">\n    <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white&style=flat-square" alt="pre-commit">\n  </a>\n</p>\n<p align="center">\n  <a href="https://pypi.org/project/segal/">\n    <img src="https://img.shields.io/pypi/v/segal.svg?logo=python&logoColor=fff&style=flat-square" alt="PyPI Version">\n  </a>\n  <img src="https://img.shields.io/pypi/pyversions/segal.svg?style=flat-square&logo=python&amp;logoColor=fff" alt="Supported Python versions">\n  <img src="https://img.shields.io/pypi/l/segal.svg?style=flat-square" alt="License">\n</p>\n\nSegAL is an active learning freamwork for semantice segmentation.\n\n## Installation\n\nSegAL is available on PyPI:\n\n`pip install segal`\n\nSegAL officially supports Python 3.8.\n\n## Active Learning Cycle\n\nTo understand what SegAL can do, we first introduce the pool-based active learning cycle.\n\n![al_cycle](./docs/images/al_cycle.png)\n\n- Step 0: Prepare seed data (a small number of labeled data used for training)\n- Step 1: Train the model with seed data\n  - Step 2: Predict unlabeled data with the trained model\n  - Step 3: Query informative samples based on predictions\n  - Step 4: Annotator (Oracle) annotate the selected samples\n  - Step 5: Input the new labeled samples to labeled dataset\n  - Step 6: Retrain model\n- Repeat step2~step6 until the f1 score of the model beyond the threshold or annotation budget is no left\n\nSegAL can simulate the whole active learning cycle.\n\n## Usage\n\nThe user can execute the below command to run the active learning cycle.\n\n```\npython examples/run_al_cycle.py --dataset CamVid  --data_path ./data/CamVid/ --model_name Unet --encoder resnet34 --encoder_weights imagenet --num_classes 12 --strategy LeastConfidence --seed_ratio 0.02 --query_ratio 0.02 --n_epoch 1\n```\n\n- `dataset`: which dataset to use\n- `data_path`: the path where the data store\n- `model_name`: name of segmentation model. More model names can be found in [architectures](https://github.com/qubvel/segmentation_models.pytorch#architectures-)\n- `encoder`: name of encoder used in model. More encoder names can be found in [encoders](https://github.com/qubvel/segmentation_models.pytorch#encoders-)\n- `encoder_weights`: pretrained weights. See [encoder table](https://github.com/qubvel/segmentation_models.pytorch#encoders-) with available weights for each encoder\n- `num_classes`: number of classes\n- `strategy`: name of sampling strategy. Available strategies: `RandomSampling`, `LeastConfidence`, `MarginSampling`, `EntropySampling`. You can find the papers for these strategy in [here](https://github.com/cure-lab/deep-active-learning/tree/main#deep-active-learning-strategies)\n- `seed_ratio`: percentage of seed data. The  used for initial training. \n- `query_ratio`: percentage of queried data in each round\n- `n_epoch`: number of epoch in each round\n\nMore explanation can be found in [usage](./docs/usage.md)\n',
    'author': 'Xu Liang',
    'author_email': 'liangxu006@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tech-sketch/SegAL',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



from .entropy_sampling import EntropySampling  # noqa: F401
from .least_confidence import LeastConfidence  # noqa: F401
from .margin_sampling import MarginSampling  # noqa: F401
from .random_sampling import RandomSampling  # noqa: F401
from .strategy import Strategy  # noqa: F401
